import logging

# import oic
from oic.extension.client import Client as ClientExtension
from oic.oic.consumer import Consumer
from oic.utils.authn.client import CLIENT_AUTHN_METHOD

from django_pyoidc.session import OIDCCacheSessionBackendForDjango
from django_pyoidc.utils import OIDCCacheBackendForDjango, get_setting_for_sso_op

logger = logging.getLogger(__name__)


class OIDCClient:
    def __init__(self, op_name, session_id=None):
        self._op_name = op_name

        self.session_cache_backend = OIDCCacheSessionBackendForDjango(self._op_name)
        self.general_cache_backend = OIDCCacheBackendForDjango(self._op_name)

        consumer_config = {
            # "debug": True,
            "response_type": "code"
        }

        client_config = {
            "client_id": get_setting_for_sso_op(op_name, "OIDC_CLIENT_ID"),
            "client_authn_method": CLIENT_AUTHN_METHOD,
        }

        self.consumer = Consumer(
            session_db=self.session_cache_backend,
            consumer_config=consumer_config,
            client_config=client_config,
        )
        # used in token introspection
        self.client_extension = ClientExtension(**client_config)

        provider_info_uri = get_setting_for_sso_op(
            op_name, "OIDC_PROVIDER_DISCOVERY_URI"
        )
        client_secret = get_setting_for_sso_op(op_name, "OIDC_CLIENT_SECRET")
        self.client_extension.client_secret = client_secret

        if session_id:
            self.consumer.restore(session_id)
        else:

            cache_key = self.general_cache_backend.generate_hashed_cache_key(
                provider_info_uri
            )
            try:
                config = self.general_cache_backend[cache_key]
            except KeyError:
                config = self.consumer.provider_config(provider_info_uri)
                # shared microcache for provider config
                # FIXME: Setting for duration
                self.general_cache_backend.set(cache_key, config, 60)
            self.consumer.client_secret = client_secret



import hashlib
import logging
from importlib import import_module
from typing import Any, Dict, Union

from django.conf import settings
from django.core.cache import BaseCache, caches

from django_pyoidc.exceptions import ClaimNotFoundError

logger = logging.getLogger(__name__)


def get_setting_for_sso_op(op_name: str, key: str, default=None):
    if key in settings.DJANGO_PYOIDC[op_name]:
        return settings.DJANGO_PYOIDC[op_name][key]
    else:
        return default


def get_settings_for_sso_op(op_name: str):
    return settings.DJANGO_PYOIDC[op_name]


def import_object(path, def_name):
    try:
        mod, cls = path.split(":", 1)
    except ValueError:
        mod = path
        cls = def_name

    return getattr(import_module(mod), cls)


def extract_claim_from_tokens(claim: str, tokens: dict) -> Any:
    """Given a dictionnary of tokens claims, extract the given claim.

    This function will seek in "info_token_claims", then "id_token_claims"
    and finally "access_token_claims".
    If the claim is not found a ClaimNotFoundError exception is raised.
    """
    if "info_token_claims" in tokens and claim in tokens["info_token_claims"]:
        value = tokens["info_token_claims"][claim]
    elif "id_token_claims" in tokens and claim in tokens["id_token_claims"]:
        value = tokens["id_token_claims"][claim]
    elif "access_token_claims" and claim in tokens["access_token_claims"]:
        value = tokens["access_token_claims"][claim]
    else:
        raise ClaimNotFoundError(f"{claim} not found in available OIDC tokens.")
    return value


def check_audience(client_id: str, access_token_claims: dict) -> bool:
    """Verify that the current client_id is present in 'aud' claim.

    Audences are stored in 'aud' claim.
    Audiences of an access token is a list of client_id where this token is allowed.
    When receiving an access token in 'API' bearer mode checking that your client_id
    is in the audience is a must.
    Access tokens received in 'full' mode, when this Django is the OIDC client
    managing the redirections to the SSO server may not always contain the client_id
    in 'aud'. This is the case for Keycloak for example, where the 'aud' would only
    contain 'others' client_id where this token can be used, and not the one generating it.
    Audience in userinfo and id tokens are different beasts.
    """
    if "aud" not in access_token_claims:
        return False
    if client_id not in access_token_claims["aud"]:
        logger.error(
            f"{client_id} not found in access_token_claims['aud']: {access_token_claims['aud']}"
        )
        return False
    return True


class OIDCCacheBackendForDjango:
    """Implement General cache for OIDC using django cache"""

    def __init__(self, op_name):
        self.op_name = op_name
        self.enabled = get_setting_for_sso_op(
            self.op_name, "OIDC_CACHE_PROVIDER_METADATA", False
        )
        if self.enabled:
            self.storage: BaseCache = caches[
                get_setting_for_sso_op(self.op_name, "CACHE_DJANGO_BACKEND")
            ]

    def generate_hashed_cache_key(self, value: str) -> str:
        h = hashlib.new("sha256")
        h.update(value.encode())
        cache_key = h.hexdigest()
        return cache_key

    def clear(self):
        return self.storage.clear()

    def get_key(self, key):
        return f"oidc-{self.op_name}-{key}"

    def set(self, key: str, value: Dict[str, Union[str, bool]], expiry: int) -> None:
        if self.enabled:
            self.storage.set(self.get_key(key), value, expiry)

    def __getitem__(self, key: str) -> Dict[str, Union[str, bool]]:
        if self.enabled:
            data = self.storage.get(self.get_key(key))
            if data is None:
                raise KeyError  # Makes __getItem__ handle like Python dict
            return data
        else:
            raise KeyError



import functools
import logging

from django.conf import settings
from django.core.exceptions import PermissionDenied
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication

from django_pyoidc.client import OIDCClient
from django_pyoidc.engine import OIDCEngine
from django_pyoidc.utils import (
    OIDCCacheBackendForDjango,
    check_audience,
    get_setting_for_sso_op,
)

logger = logging.getLogger(__name__)


class OidcAuthException(Exception):
    pass


class OIDCBearerAuthentication(BaseAuthentication):
    def __init__(self, *args, **kwargs):
        super(OIDCBearerAuthentication, self).__init__(*args, **kwargs)
        self.op_name = self.extract_drf_opname()
        self.general_cache_backend = OIDCCacheBackendForDjango(self.op_name)
        self.engine = OIDCEngine(self.op_name)

    @functools.cached_property
    def client(self):
        return OIDCClient(self.op_name)

    @classmethod
    def extract_drf_opname(cls):
        """
        Given a list of opnames and setting in DJANGO_PYOIDC conf, extract the one having USED_BY_REST_FRAMEWORK=True.
        """
        op = None
        found = False
        for op_name, configs in settings.DJANGO_PYOIDC.items():
            if (
                "USED_BY_REST_FRAMEWORK" in configs
                and configs["USED_BY_REST_FRAMEWORK"]
            ):
                if found:
                    raise RuntimeError(
                        "Several DJANGO_PYOIDC sections are declared as USED_BY_REST_FRAMEWORK, only one should be used."
                    )
                found = True
                op = op_name
        if found:
            return op
        else:
            raise RuntimeError(
                "No DJANGO_PYOIDC sections are declared with USED_BY_REST_FRAMEWORK configuration option."
            )

    def extract_access_token(self, request) -> str:
        val = request.headers.get("Authorization")
        if not val:
            msg = "Request missing the authorization header."
            raise OidcAuthException(msg)
        val = val.strip()
        bearer_name, access_token_jwt = val.split(maxsplit=1)
        requested_bearer_name = get_setting_for_sso_op(
            self.op_name, "OIDC_API_BEARER_NAME", "Bearer"
        )
        if not bearer_name.lower() == requested_bearer_name.lower():
            msg = f"Bad authorization header, invalid Keyword for the bearer, expecting {requested_bearer_name}."
            raise OidcAuthException(msg)
        return access_token_jwt

    def authenticate(self, request):
        """
        Returns two-tuple of (user, token) if authentication succeeds,
        or None otherwise.
        """
        try:
            user = None
            access_token_claims = None

            # Extract the access token from an HTTP Authorization Bearer header
            try:
                access_token_jwt = self.extract_access_token(request)
            except OidcAuthException as e:
                logger.debug(e)
                return None

            # This introspection of the token is made by the SSO server
            # so it is quite slow, but there's a cache added based on the token expiration
            access_token_claims = self.engine.introspect_access_token(
                access_token_jwt, client=self.client
            )
            logger.debug(access_token_claims)
            if not access_token_claims.get("active"):
                msg = "Inactive access token."
                raise exceptions.AuthenticationFailed(msg)

            # FIXME: add an option to request userinfo here, but that may be quite slow

            if access_token_claims:
                logger.debug("Request has valid access token.")

                # FIXME: Add a setting to disable
                client_id = get_setting_for_sso_op(self.op_name, "OIDC_CLIENT_ID")
                if not check_audience(client_id, access_token_claims):
                    raise PermissionDenied(
                        f"Invalid result for acces token audiences check for {client_id}."
                    )

                logger.debug("Let application load user via user hook.")
                user = self.engine.call_get_user_function(
                    tokens={
                        "access_token_jwt": access_token_jwt,
                        "access_token_claims": access_token_claims,
                    }
                )

            if not user:
                # OIDC Bearer Authentication process failure : cannot set active authenticated user
                return None

        except PermissionDenied as exp:
            raise exp
        except Exception as exp:
            logger.exception(exp)
            return None

        return user, access_token_claims



import logging
from urllib.parse import urljoin

logger = logging.getLogger(__name__)

try:
    from drf_spectacular.extensions import OpenApiAuthenticationExtension

    from django_pyoidc.utils import get_setting_for_sso_op

    class OIDCScheme(OpenApiAuthenticationExtension):
        target_class = "django_pyoidc.drf.authentication.OIDCBearerAuthentication"
        name = "openIdConnect"
        match_subclasses = True
        priority = -1

        def get_security_definition(self, auto_schema):
            from django_pyoidc.drf.authentication import OIDCBearerAuthentication

            op = OIDCBearerAuthentication.extract_drf_opname()
            well_known_url = get_setting_for_sso_op(op, "OIDC_PROVIDER_DISCOVERY_URI")
            if not well_known_url.endswith(".well-known/openid-configuration"):
                if not well_known_url.endswith("/"):
                    well_known_url += "/"
                well_known_url = urljoin(
                    well_known_url, ".well-known/openid-configuration"
                )

            header_name = get_setting_for_sso_op(op, "OIDC_API_BEARER_NAME", "Bearer")
            if header_name != "Bearer":
                logger.warning(
                    "The configuration for 'OIDC_API_BEARER_NAME' will cause issue with swagger UI :"
                    "it is not yet possible to change the header name for swagger UI, you should stick to"
                    "'Bearer'."
                )
            return {
                "type": "openIdConnect",
                "openIdConnectUrl": well_known_url,
            }

except ImportError:
    logger.debug(
        "The package 'drf-spectacular' is not installed, skipping schema generation."
    )



import datetime
import logging

from django_pyoidc import get_user_by_email
from django_pyoidc.client import OIDCClient
from django_pyoidc.utils import (
    OIDCCacheBackendForDjango,
    get_setting_for_sso_op,
    import_object,
)

logger = logging.getLogger(__name__)


class OIDCEngine:
    def __init__(self, op_name: str):
        self.op_name = op_name
        self.general_cache_backend = OIDCCacheBackendForDjango(self.op_name)

    def call_function(self, setting_name, *args, **kwargs):
        function_path = get_setting_for_sso_op(self.op_name, setting_name)
        if function_path:
            func = import_object(function_path, "")
            return func(*args, **kwargs)

    def call_get_user_function(self, tokens={}):
        if get_setting_for_sso_op(self.op_name, "HOOK_GET_USER"):
            logger.debug("OIDC, Calling user hook on get_user")
            return self.call_function("HOOK_GET_USER", tokens)
        else:
            return get_user_by_email(tokens)

    def introspect_access_token(self, access_token_jwt: str, client: OIDCClient):
        """
        Perform a cached intropesction call to extract claims from encoded jwt of the access_token
        """
        # FIXME: allow a non-cached mode by global settings
        access_token_claims = None

        # FIXME: in what case could we not have an access token available?
        # should we raise an error then?
        if access_token_jwt is not None:
            cache_key = self.general_cache_backend.generate_hashed_cache_key(
                access_token_jwt
            )
            try:
                access_token_claims = self.general_cache_backend["cache_key"]
            except KeyError:
                # CACHE MISS

                # RFC 7662: token introspection: ask SSO to validate and render the jwt as json
                # this means a slow web call
                request_args = {
                    "token": access_token_jwt,
                    "token_type_hint": "access_token",
                }
                client_auth_method = client.consumer.registration_response.get(
                    "introspection_endpoint_auth_method", "client_secret_basic"
                )
                introspection = client.client_extension.do_token_introspection(
                    request_args=request_args,
                    authn_method=client_auth_method,
                    endpoint=client.consumer.introspection_endpoint,
                )
                access_token_claims = introspection.to_dict()

                # store it in cache
                current = datetime.datetime.now().strftime("%s")
                if "exp" not in access_token_claims:
                    raise RuntimeError("No expiry set on the access token.")
                access_token_expiry = access_token_claims["exp"]
                exp = int(access_token_expiry) - int(current)
                logger.debug(
                    f"Token expiry: {exp}  - current is {current} "
                    f"and expiry is set to {access_token_expiry} in the token"
                )
                self.general_cache_backend.set(cache_key, access_token_claims, exp)
        return access_token_claims
